"""Delta E 94."""
import math
from ... import util


def distance(color1, color2, kl=1, k1=0.045, k2=0.015):
    """
    Delta E 1994 color distance formula.

    http://www.brucelindbloom.com/Eqn_DeltaE_CIE94.html
    """

    l1, a1, b1 = util.no_nan(color1.convert("lab").coords())
    l2, a2, b2 = util.no_nan(color2.convert("lab").coords())

    # Equation (5)
    c1 = math.sqrt(a1 ** 2 + b1 ** 2)

    # Equation (6)
    c2 = math.sqrt(a2 ** 2 + b2 ** 2)

    # Equation  (2)
    dl = l1 - l2

    # Equation  (3)
    dc = c1 - c2

    # Equation (7)
    da = a1 - a2

    # Equation  (8)
    db = b1 - b2

    # Equation (4)
    # We never reference `dh` until the very end, and when we do, we square it
    # before using it, so we don't need the square root as described in the
    # algorithm. Instead we can just leave the result as is.
    dh = da ** 2 + db ** 2 - dc ** 2

    # Equation (9)
    sl = 1

    # Equation (10)
    sc = 1 + k1 * c1

    # Equation (11)
    sh = 1 + k2 * c1

    # Equation (12)
    # Provided by `kl`

    # Equation (13)
    kc = 1

    # Equation (14)
    kh = 1

    # Equation (15) and Equation (16)
    # Provided by `k1` and `k2`

    # Equation (1)
    return math.sqrt(
        (dl / (kl * sl)) ** 2 +
        (dc / (kc * sc)) ** 2 +
        # Square root just the denominator as `dh` is already squared.
        dh / ((kh * sh) ** 2)
    )



"""CSS Color object."""
from .spaces.srgb import SRGB
from .spaces.hsl import HSL
from .spaces.hwb import HWB
from .spaces.lab import Lab
from .spaces.lch import Lch
from ..color import Color as GenericColor

CSS_OVERRIDES = (HSL, HWB, Lab, Lch, SRGB)


class Color(GenericColor):
    """Color wrapper class."""

    CS_MAP = {key: value for key, value in GenericColor.CS_MAP.items()}
    for color in CSS_OVERRIDES:
        CS_MAP[color.space()] = color



"""Euclidean distance."""
import math
from ... import util


def distance(color1, color2, space="lab", **kwargs):
    """
    Euclidean distance.

    https://en.wikipedia.org/wiki/Euclidean_distance
    """

    coords1 = util.no_nan(color1.convert(space).coords())
    coords2 = util.no_nan(color2.convert(space).coords())

    return math.sqrt(sum((x - y) ** 2.0 for x, y in zip(coords2, coords1)))



"""Clip to fit in gamut."""
from ... import util
from ... spaces import Angle, GamutBound


def fit(color):
    """Gamut clipping."""

    channels = util.no_nan(color.coords())
    gamut = color._space.RANGE
    fit = []

    for i, value in enumerate(channels):
        a, b = gamut[i]
        is_bound = isinstance(gamut[i], GamutBound)

        # Wrap the angle. Not technically out of gamut, but we will clean it up.
        if isinstance(a, Angle) and isinstance(b, Angle):
            fit.append(value % 360.0)
            continue

        # These parameters are unbounded
        if not is_bound:  # pragma: no cover
            # Will not execute unless we have a space that defines some coordinates
            # as bound and others as not. We do not currently have such spaces.
            a = None
            b = None

        # Fit value in bounds.
        fit.append(util.clamp(value, a, b))
    return fit



"""Delta E 2000."""
import math
from ... import util

G_CONST = 25 ** 7


def distance(color1, color2, kl=1, kc=1, kh=1, **kwargs):
    """
    Calculate distance doing a direct translation of the algorithm from the CIE Delta E 2000 paper.

    We denoted prime (L') with trailing 'p' and mean is represented with a trailing 'm'.
    Delta has a preceding 'd'. I'm not sure I was completely consistent.

    http://www2.ece.rochester.edu/~gsharma/ciede2000/ciede2000noteCRNA.pdf
    """

    l1, a1, b1 = util.no_nan(color1.convert("lab").coords())
    l2, a2, b2 = util.no_nan(color2.convert("lab").coords())

    # Equation (2)
    c1 = math.sqrt(a1 ** 2 + b1 ** 2)
    c2 = math.sqrt(a2 ** 2 + b2 ** 2)

    # Equation (3)
    cm = (c1 + c2) / 2

    # Equation (4)
    c7 = cm ** 7
    g = 0.5 * (1 - math.sqrt(c7 / (c7 + G_CONST)))

    # Equation (5)
    ap1 = (1 + g) * a1
    ap2 = (1 + g) * a2

    # Equation (6)
    cp1 = math.sqrt(ap1 ** 2 + b1 ** 2)
    cp2 = math.sqrt(ap2 ** 2 + b2 ** 2)

    # Equation (7)
    hp1 = 0 if (ap1 == 0 and b1 == 0) else math.atan2(b1, ap1)
    hp2 = 0 if (ap2 == 0 and b2 == 0) else math.atan2(b2, ap2)
    hp1 = math.degrees(hp1 + 2 * math.pi if hp1 < 0.0 else hp1)
    hp2 = math.degrees(hp2 + 2 * math.pi if hp2 < 0.0 else hp2)

    # Equation (8)
    dl = l2 - l1

    # Equation (9)
    dc = cp2 - cp1

    # Equation (10)
    hdiff = hp2 - hp1
    if cp1 * cp2 == 0.0:
        dh = 0.0
    elif abs(hdiff) <= 180.0:
        dh = hdiff
    else:
        # If not `hdiff > 180.0` and not `abs(hdiff) <= 180.0`
        # then it must be `hdiff < -180`
        offset = -360 if hdiff > 180.0 else 360
        dh = hdiff + offset

    # Equation (11)
    dh = 2 * math.sqrt(cp2 * cp1) * math.sin(math.radians(dh / 2))

    # Equation (12)
    lpm = (l1 + l2) / 2

    # Equation (13)
    cpm = (cp1 + cp2) / 2

    # Equation (14)
    hsum = hp1 + hp2
    if cp1 * cp2 == 0:
        hpm = hsum
    elif abs(hp1 - hp2) > 180:
        # if not `hsum < 360`
        # then it must be `hsum >= 360`
        offset = 360 if hsum < 360 else -360
        hpm = (hsum + offset) / 2
    else:  # `abs(hp1 - hp2) <= 180`
        hpm = hsum / 2

    # Equation (15)
    t = (
        1 -
        (0.17 * math.cos(math.radians(hpm - 30))) +
        (0.24 * math.cos(math.radians(2 * hpm))) +
        (0.32 * math.cos(math.radians((3 * hpm) + 6))) -
        (0.20 * math.cos(math.radians((4 * hpm) - 63)))
    )

    # Equation (16)
    dt = 30 * math.exp(-1 * ((hpm - 275) / 25) ** 2)

    # Equation (17)
    cpm7 = cpm ** 7
    rc = 2 * math.sqrt(cpm7 / (cpm7 + G_CONST))

    # Equation (18)
    l_temp = (lpm - 50) ** 2
    sl = 1 + ((0.015 * l_temp) / math.sqrt(20 + l_temp))

    # Equation (19)
    sc = 1 + 0.045 * cpm

    # Equation (20)
    sh = 1 + 0.015 * cpm * t

    # Equation (21)
    rt = -1 * math.sin(math.radians(2 * dt)) * rc

    # Equation (22)
    return math.sqrt(
        (dl / (kl * sl)) ** 2 +
        (dc / (kc * sc)) ** 2 +
        (dh / (kh * sh)) ** 2 +
        rt * (dc / (kc * sc)) * (dh / (kh * sh))
    )



"""Delta E 76."""

from . import distance_euclidean


def distance(color1, color2, **kwargs):
    """
    Delta E 1976 color distance formula.

    http://www.brucelindbloom.com/index.html?Eqn_DeltaE_CIE76.html

    Basically this is Euclidean distance in the Lab space.
    """

    # Equation (1)
    return distance_euclidean.distance(color1, color2, space="lab")



"""Match input colors."""


class ColorMatch:
    """Color match object."""

    def __init__(self, color, start, end):
        """Initialize."""

        self.color = color
        self.start = start
        self.end = end

    def __str__(self):  # pragma: no cover
        """String."""

        return "ColorMatch(color={!r}, start={}, end={})".format(self.color, self.start, self.end)

    __repr__ = __str__


class Match:
    """Match support."""

    @classmethod
    def _match(cls, string, start=0, fullmatch=False, filters=None):
        """
        Match a color in a buffer and return a color object.

        This must return the color space, not the Color object.
        """

        filters = set(filters) if filters is not None else set()

        for space, space_class in cls.CS_MAP.items():
            if filters and space not in filters:
                continue
            value, match_end = space_class.match(string, start, fullmatch)
            if value is not None:
                color = space_class(*value)
                return ColorMatch(color, start, match_end)
        return None

    @classmethod
    def match(cls, string, start=0, fullmatch=False, *, filters=None):
        """Match color."""

        obj = cls._match(string, start, fullmatch, filters=filters)
        if obj is not None:
            obj.color = cls(obj.color.space(), obj.color.coords(), obj.color.alpha)
        return obj



"""
Interpolation methods.

A lot of code was ported and or adapted from the https://colorjs.io project. Particularly
the `interpolate` method and the functions built on top of it, such as `mix` and `steps`.

While we deviate in some ways, a lot of it, at the time of this comment, are a direct port.

In general, the logic mimics in many ways the `color-mix` function as outlined in the Level 5
color draft (Oct 2020), but the approach was modeled directly off of the work done in color.js.
---
Original Authors: Lea Verou, Chris Lilley
License: MIT (As noted in https://github.com/LeaVerou/color.js/blob/master/package.json)
"""
import math
from abc import ABCMeta, abstractmethod
from collections.abc import Sequence, Mapping, Callable
from collections import namedtuple
from .. import util
from ..spaces import Cylindrical, Angle


class Lerp:
    """Linear interpolation."""

    def __init__(self, progress):
        """Initialize."""

        self.progress = progress

    def __call__(self, a, b, t):
        """Interpolate with period."""

        return a + (b - a) * (t if not isinstance(self.progress, Callable) else self.progress(t))


class Piecewise(namedtuple('Piecewise', ['color', 'stop', 'progress', 'hue', 'premultiplied'])):
    """Piecewise interpolation input."""

    __slots__ = ()

    def __new__(cls, color, stop=None, progress=None, hue=util.DEF_HUE_ADJ, premultiplied=False):
        """Initialize."""

        return super().__new__(cls, color, stop, progress, hue, premultiplied)


class Interpolator(metaclass=ABCMeta):
    """Interpolator."""

    @abstractmethod
    def __init__(self):
        """Initialize."""

    @abstractmethod
    def __call__(self, p):
        """Call the interpolator."""

    @abstractmethod
    def get_delta(self):
        """Initialize."""

    def steps(self, steps=2, max_steps=1000, max_delta_e=0):
        """Steps."""

        return color_steps(self, steps, max_steps, max_delta_e)


class InterpolateSingle(Interpolator):
    """Interpolate a single range of two colors."""

    def __init__(self, channels1, channels2, names, create, progress, space, outspace, premultiplied):
        """Initialize."""

        self.names = names
        self.channels1 = channels1
        self.channels2 = channels2
        self.create = create
        self.progress = progress
        self.space = space
        self.outspace = outspace
        self.premultiplied = premultiplied

    def get_delta(self):
        """Get the delta."""

        return self.create(self.space, self.channels1).delta_e(self.create(self.space, self.channels2))

    def __call__(self, p):
        """Run through the coordinates and run the interpolation on them."""

        channels = []
        for i, c1 in enumerate(self.channels1):
            name = self.names[i]
            c2 = self.channels2[i]
            if util.is_nan(c1) and util.is_nan(c2):
                value = 0.0
            elif util.is_nan(c1):
                value = c2
            elif util.is_nan(c2):
                value = c1
            else:
                progress = None
                if isinstance(self.progress, Mapping):
                    progress = self.progress.get(name, self.progress.get('all'))
                else:
                    progress = self.progress
                lerp = progress if isinstance(progress, Lerp) else Lerp(progress)
                value = lerp(c1, c2, p)
            channels.append(value)
        color = self.create(self.space, channels[:-1], channels[-1])
        if self.premultiplied:
            postdivide(color)
        return color.convert(self.outspace, in_place=True) if self.outspace != color.space() else color


class InterpolatePiecewise(Interpolator):
    """Interpolate multiple ranges of colors."""

    def __init__(self, stops, interpolators):
        """Initialize."""

        self.start = stops[0]
        self.end = stops[len(stops) - 1]
        self.stops = stops
        self.interpolators = interpolators

    def get_delta(self):
        """Get the delta total."""

        return [i.get_delta() for i in self.interpolators]

    def __call__(self, p):
        """Interpolate."""

        percent = p
        if percent > self.end:
            # Beyond range, just interpolate the last colors
            return self.interpolators[-1](1 + abs(p - self.end) if p > 1 else 1)

        elif percent < self.start:
            # Beyond range, just interpolate the last colors
            return self.interpolators[0](0 - abs(self.start - p) if p < 0 else 0)

        else:
            last = self.start
            for i, interpolator in enumerate(self.interpolators, 1):
                stop = self.stops[i]
                if percent <= stop:
                    r = stop - last
                    p2 = (percent - last) / r if r else 1
                    return interpolator(p2)
                last = stop


def calc_stops(stops, count):
    """Calculate stops."""

    # Ensure the first stop is set to zero if not explicitly set
    if 0 not in stops:
        stops[0] = 0

    last = stops[0] * 100
    highest = last
    empty = None
    final = {}

    # Build up normalized stops
    for i in range(count):
        value = stops.get(i)
        if value is not None:
            value *= 100

        # Found an empty hole, track the start
        if value is None and empty is None:
            empty = i - 1
            continue
        elif value is None:
            continue

        # We can't have a stop decrease in progression
        if value < last:
            value = last

        # Track the largest explicit value set
        if value > highest:
            highest = value

        # Fill in hole if one exists.
        # Holes will be evenly space between the
        # current and last stop.
        if empty is not None:
            r = i - empty
            increment = (value - last) / r
            for j in range(empty + 1, i):
                last += increment
                final[j] = last / 100
            empty = None

        # Set the stop and track it as the last
        last = value
        final[i] = last / 100

    # If there is a hole at the end, fill in the hole,
    # equally spacing the stops from the last to 100%.
    # If the last is greater than 100%, then all will
    # be equal to the last.
    if empty is not None:
        r = (count - 1) - empty
        if highest > 100:
            increment = 0
        else:
            increment = (100 - last) / r
        for j in range(empty + 1, count):
            last += increment
            final[j] = last / 100

    return final


def postdivide(color):
    """Premultiply the given transparent color."""

    if color.alpha >= 1.0:
        return

    channels = color.coords()
    gamut = color._space.RANGE
    alpha = color.alpha
    coords = []
    for i, value in enumerate(channels):
        a = gamut[i][0]

        # Wrap the angle
        if isinstance(a, Angle):
            coords.append(value)
            continue
        coords.append(value / alpha if alpha != 0 else value)
    color._space._coords = coords


def premultiply(color):
    """Premultiply the given transparent color."""

    if color.alpha >= 1.0:
        return

    channels = color.coords()
    gamut = color._space.RANGE
    alpha = color.alpha
    coords = []
    for i, value in enumerate(channels):
        a = gamut[i][0]

        # Wrap the angle
        if isinstance(a, Angle):
            coords.append(value)
            continue
        coords.append(value * alpha)
    color._space._coords = coords


def adjust_hues(color1, color2, hue):
    """Adjust hues."""

    hue = hue.lower()
    if hue == "specified":
        return

    name = color1._space.hue_name()
    c1 = color1.get(name)
    c2 = color2.get(name)

    c1 = c1 % 360
    c2 = c2 % 360

    if util.is_nan(c1) or util.is_nan(c2):
        color1.set(name, c1)
        color2.set(name, c2)
        return

    if hue == "shorter":
        if c2 - c1 > 180:
            c1 += 360
        elif c2 - c1 < -180:
            c2 += 360

    elif hue == "longer":
        if 0 < (c2 - c1) < 180:
            c1 += 360
        elif -180 < (c2 - c1) < 0:
            c2 += 360

    elif hue == "increasing":
        if c2 < c1:
            c2 += 360

    elif hue == "decreasing":
        if c1 < c2:
            c1 += 360

    else:
        raise ValueError("Unknown hue adjuster '{}'".format(hue))

    color1.set(name, c1)
    color2.set(name, c2)


def color_steps(interpolator, steps=2, max_steps=1000, max_delta_e=0):
    """Color steps."""

    if max_delta_e <= 0:
        actual_steps = steps
    else:
        actual_steps = 0
        deltas = interpolator.get_delta()
        if not isinstance(deltas, Sequence):
            deltas = [deltas]
        actual_steps = sum([d / max_delta_e for d in deltas])
        actual_steps = max(steps, math.ceil(actual_steps) + 1)

    if max_steps is not None:
        actual_steps = min(actual_steps, max_steps)

    ret = []
    if actual_steps == 1:
        ret = [{"p": 0.5, "color": interpolator(0.5)}]
    else:
        step = 1 / (actual_steps - 1)
        for i in range(actual_steps):
            p = i * step
            ret.append({'p': p, 'color': interpolator(p)})

    # Iterate over all the stops inserting stops in between if all colors
    # if we have any two colors with a max delta greater than what was requested.
    # We inject between every stop to ensure the midpoint does not shift.
    if max_delta_e > 0:
        # Initial check to see if we need to insert more stops
        m_delta = 0
        for i, entry in enumerate(ret):
            if i == 0:
                continue
            m_delta = max(m_delta, entry['color'].delta_e(ret[i - 1]['color']))

        while m_delta > max_delta_e:
            # Inject stops while measuring again to see if it was sufficient
            m_delta = 0
            i = 1
            while i < len(ret) and len(ret) < max_steps:
                prev = ret[i - 1]
                cur = ret[i]
                p = (cur['p'] + prev['p']) / 2
                color = interpolator(p)
                m_delta = max(m_delta, color.delta_e(prev['color']), color.delta_e(cur['color']))
                ret.insert(i, {'p': p, 'color': color})
                i += 2

    return [i['color'] for i in ret]


def color_piecewise_lerp(pw, space, out_space, progress, hue, premultiplied):
    """Piecewise Interpolation."""

    # Ensure we have something we can interpolate with
    count = len(pw)
    if count == 1:
        pw = [pw[0], pw[0]]
        count += 1

    # Calculate stops
    stops = {}
    for i, x in enumerate(pw, 0):
        if not isinstance(x, Piecewise):
            pw[i] = Piecewise(x)
        elif x.stop is not None:
            stops[i] = x.stop
    stops = calc_stops(stops, count)

    # Construct piecewise interpolation object
    color_map = []
    current = pw[0].color
    for i in range(1, count):
        p = pw[i]
        color = current._handle_color_input(p.color)

        color_map.append(
            current.interpolate(
                color,
                space=space,
                out_space=out_space,
                progress=p.progress if p.progress is not None else progress,
                hue=p.hue if p.hue is not None else hue,
                premultiplied=p.premultiplied if p.premultiplied is not None else premultiplied
            )
        )
        current = color

    return InterpolatePiecewise(stops, color_map)


def color_lerp(color1, color2, space, out_space, progress, hue, premultiplied):
    """Color interpolation."""

    # Convert to the color space and ensure the color fits inside
    color1 = color1.convert(space, fit=True)
    color2 = color1._handle_color_input(color2).convert(space, fit=True)

    # Adjust hues if we have two valid hues
    if isinstance(color1._space, Cylindrical):
        adjust_hues(color1, color2, hue)

    if premultiplied:
        premultiply(color1)
        premultiply(color2)

    channels1 = color1.coords()
    channels2 = color2.coords()

    # Include alpha
    channels1.append(color1.alpha)
    channels2.append(color2.alpha)

    return InterpolateSingle(
        names=color1._space.CHANNEL_NAMES,
        channels1=channels1,
        channels2=channels2,
        create=type(color1),
        progress=progress,
        space=space,
        outspace=out_space,
        premultiplied=premultiplied
    )


class Interpolate:
    """Interpolate between colors."""

    def mask(self, channel, *, invert=False, in_place=False):
        """Mask color channels."""

        this = self if in_place else self.clone()
        masks = set([channel] if isinstance(channel, str) else channel)
        for name in self._space.CHANNEL_NAMES:
            if (not invert and name in masks) or (invert and name not in masks):
                this.set(name, util.NaN)
        return this

    def steps(self, color, *, steps=2, max_steps=1000, max_delta_e=0, **interpolate_args):
        """
        Discrete steps.

        This is built upon the interpolate function, and will return a list of
        colors containing a minimum of colors equal to `steps` or steps as specified
        derived from the `max_delta_e` parameter (whichever is greatest).

        Number of colors can be capped with `max_steps`.

        Default delta E method used is delta E 76.
        """

        return self.interpolate(color, **interpolate_args).steps(steps, max_steps, max_delta_e)

    def mix(self, color, percent=util.DEF_MIX, *, in_place=False, **interpolate_args):
        """
        Mix colors using interpolation.

        This uses the interpolate method to find the center point between the two colors.
        The basic mixing logic is outlined in the CSS level 5 draft.
        """

        if not self._is_color(color) and not isinstance(color, (str, Piecewise)):
            raise TypeError("Unexpected type '{}'".format(type(color)))
        color = self.interpolate(color, **interpolate_args)(percent)
        return self.mutate(color) if in_place else color

    def interpolate(
        self, color, *, space="lab", out_space=None, stop=0, progress=None, hue=util.DEF_HUE_ADJ, premultiplied=False
    ):
        """
        Return an interpolation function.

        The function will return an interpolation function that accepts a value (which should
        be in the range of [0..1] and will return a color based on that value.

        While we use NaNs to mask off channels when doing the interpolation, we do not allow
        arbitrary specification of NaNs by the user, they must specify channels via `adjust`
        if they which to target specific channels for mixing. Null hues become NaNs before
        mixing occurs.
        """

        space = space.lower()
        out_space = self.space() if out_space is None else out_space.lower()

        # A piecewise object was provided, so treat it as such,
        # or we've changed the stop of the base color, so run it through piecewise.
        if (
            isinstance(color, Piecewise) or
            (stop != 0 and (isinstance(color, str) or self._is_color(color)))
        ):
            color = [color]

        if not isinstance(color, str) and isinstance(color, Sequence):
            # We have a sequence, so use piecewise interpolation
            return color_piecewise_lerp(
                [Piecewise(self, stop=stop)] + list(color),
                space,
                out_space,
                progress,
                hue,
                premultiplied
            )
        else:
            # We have a sequence, so use piecewise interpolation
            return color_lerp(
                self,
                color,
                space,
                out_space,
                progress,
                hue,
                premultiplied
            )



"""Gamut handling."""
from ... import util
from ... spaces import Angle, GamutBound
from . import clip
from . import lch_chroma


def norm_angles(color):
    """Normalize angles."""

    channels = util.no_nan(color.coords())
    gamut = color._space.RANGE
    fit = []
    for i, value in enumerate(channels):
        a = gamut[i][0]

        # Wrap the angle
        if isinstance(a, Angle):
            fit.append(value % 360.0)
            continue

        # Fit value in bounds.
        fit.append(value)
    return fit


class Gamut:
    """Handle gamut related functions."""

    def fit(self, space=None, *, method=None, in_place=False):
        """Fit the gamut using the provided method."""

        if space is None:
            space = self.space()

        if method is None:
            method = self.FIT

        this = self.clone() if not in_place else self

        # Select appropriate mapping algorithm
        if method == "clip":
            func = clip.fit
        elif method == "lch-chroma":
            func = lch_chroma.fit
        else:
            # Unknown fit method
            raise ValueError("'{}' gamut mapping is not currently supported".format(method))

        # Convert to desired space
        c = self.convert(space)

        # If we are perfectly in gamut, don't waste time fitting, just normalize hues.
        # If out of gamut, apply mapping/clipping/etc.
        c._space._coords = norm_angles(c) if c.in_gamut(tolerance=0.0) else func(c)

        # Adjust "this" color
        return this.update(c)

    def in_gamut(self, space=None, *, tolerance=util.DEF_FIT_TOLERANCE):
        """Check if current color is in gamut."""

        space = space.lower() if space is not None else self.space()

        # Check gamut in the provided space
        if space is not None and space != self.space():
            c = self.convert(space)
            return c.in_gamut(tolerance=tolerance)

        # Check the color space specified for gamut checking.
        # If it proves to be in gamut, we will then test if the current
        # space is constrained properly.
        if self._space.GAMUT_CHECK is not None:
            c = self.convert(self._space.GAMUT_CHECK)
            if not c.in_gamut(tolerance=tolerance):
                return False

        # Verify the values are in bound
        channels = util.no_nan(self.coords())
        for i, value in enumerate(channels):
            a, b = self._space.RANGE[i]
            is_bound = isinstance(self._space.RANGE[i], GamutBound)

            # Angles will wrap, so no sense checking them
            if isinstance(a, Angle):
                continue

            # These parameters are unbounded
            if not is_bound:
                a = None
                b = None

            # Check if bounded values are in bounds
            if (a is not None and value < (a - tolerance)) or (b is not None and value > (b + tolerance)):
                return False

        return True



"""Colors."""
from collections.abc import Sequence
from . import distance
from . import convert
from . import gamut
from . import compositing
from . import interpolate
from . import contrast
from . import match
from .. import util
from ..spaces.hsv import HSV
from ..spaces.srgb import SRGB
from ..spaces.srgb_linear import SRGBLinear
from ..spaces.hsl import HSL
from ..spaces.hwb import HWB
from ..spaces.lab import Lab
from ..spaces.lch import Lch
from ..spaces.lab_d65 import LabD65
from ..spaces.lch_d65 import LchD65
from ..spaces.display_p3 import DisplayP3
from ..spaces.a98_rgb import A98RGB
from ..spaces.prophoto_rgb import ProPhotoRGB
from ..spaces.rec2020 import Rec2020
from ..spaces.xyz import XYZ
from ..spaces.xyz_d65 import XYZD65
from ..spaces.oklab import Oklab
from ..spaces.oklch import Oklch
from ..spaces.jzazbz import Jzazbz
from ..spaces.jzczhz import JzCzhz
from ..spaces.ictcp import ICtCp


SUPPORTED = (
    HSL, HWB, Lab, Lch, LabD65, LchD65, SRGB, SRGBLinear, HSV,
    DisplayP3, A98RGB, ProPhotoRGB, Rec2020, XYZ, XYZD65,
    Oklab, Oklch, Jzazbz, JzCzhz, ICtCp
)


class Color(
    convert.Convert,
    gamut.Gamut,
    compositing.Compose,
    interpolate.Interpolate,
    distance.Distance,
    contrast.Contrast,
    match.Match
):
    """Color class object which provides access and manipulation of color spaces."""

    CS_MAP = {obj.space(): obj for obj in SUPPORTED}

    PRECISION = util.DEF_PREC
    FIT = util.DEF_FIT
    DELTA_E = util.DEF_DELTA_E
    CHROMATIC_ADAPTATION = 'bradford'

    def __init__(self, color, data=None, alpha=util.DEF_ALPHA, *, filters=None, **kwargs):
        """Initialize."""

        self._attach(self._parse(color, data, alpha, filters=filters, **kwargs))

    def __eq__(self, other):
        """Compare equal."""

        return (
            other.space() == self.space() and
            util.cmp_coords(other.coords(), self.coords()) and
            util.cmp_coords(other.alpha, self.alpha)
        )

    def _parse(self, color, data=None, alpha=util.DEF_ALPHA, filters=None, **kwargs):
        """Parse the color."""

        obj = None
        if data is not None:
            filters = set(filters) if filters is not None else set()
            for space, space_class in self.CS_MAP.items():
                s = color.lower()
                if space == s and (not filters or s in filters):
                    if len(data) < space_class.NUM_COLOR_CHANNELS:
                        data = list(data) + [util.NaN] * (space_class.NUM_COLOR_CHANNELS - len(data))
                    obj = space_class(data[:space_class.NUM_COLOR_CHANNELS], alpha)
                    return obj
        elif isinstance(color, Color):
            if not filters or color.space() in filters:
                obj = self.CS_MAP[color.space()](color._space)
        else:
            m = self._match(color, fullmatch=True, filters=filters)
            if m is None:
                raise ValueError("'{}' is not a valid color".format(color))
            obj = m.color
        if obj is None:
            raise ValueError("Could not process the provided color")
        return obj

    def is_nan(self, name):
        """Check if channel is NaN."""

        return util.is_nan(self.get(name))

    def _is_this_color(self, obj):
        """Test if the input is "this" Color, not a subclass."""

        return type(obj) is type(self)

    def _is_color(self, obj):
        """Test if the input is a Color."""

        return isinstance(obj, Color)

    def _attach(self, space):
        """Attach the this objects convert space to the color."""

        self._space = space

    def _handle_color_input(self, color, sequence=False):
        """Handle color input."""

        if isinstance(color, str) or (self._is_color(color) and not self._is_this_color(color)):
            color = self.new(color)
        elif sequence and isinstance(color, Sequence):
            color = [self._handle_color_input(c) for c in color]
        elif not self._is_color(color):
            raise TypeError("Unexpected type '{}'".format(type(color)))
        return color

    def space(self):
        """The current color space."""

        return self._space.space()

    def coords(self):
        """Coordinates."""

        return self._space.coords()

    def new(self, color, data=None, alpha=util.DEF_ALPHA, *, filters=None, **kwargs):
        """
        Create new color object.

        TODO: maybe allow `currentcolor` here? It would basically clone the current object.
        """

        return type(self)(color, data, alpha, filters=filters, **kwargs)

    def clone(self):
        """Clone."""

        return self.new(self.space(), self.coords(), self.alpha)

    def to_string(self, **kwargs):
        """To string."""

        return self._space.to_string(self, **kwargs)

    def __repr__(self):
        """Representation."""

        return repr(self._space)

    __str__ = __repr__

    def get(self, name):
        """Get channel."""

        # Handle space.attribute
        if '.' in name:
            parts = name.split('.')
            if len(parts) != 2:
                raise ValueError("Could not resolve attribute '{}'".format(name))
            obj = self.convert(parts[0])
            return obj.get(parts[1])

        return self._space.get(name)

    def set(self, name, value):  # noqa: A003
        """Set channel."""

        # Handle space.attribute
        if '.' in name:
            parts = name.split('.')
            if len(parts) != 2:
                raise ValueError("Could not resolve attribute '{}'".format(name))
            obj = self.convert(parts[0])
            obj.set(parts[1], value)
            return self.update(obj)

        # Handle a function that modifies the value or a direct value
        if callable(value):
            self.set(name, value(self.get(name)))
        else:
            self._space.set(name, value)
        return self

    def __getattr__(self, name):
        """Get attribute."""

        # Don't test `_space` as it is used to get Space channel attributes.
        if name != "_space":
            # Get channel names
            names = set()
            result = getattr(self, "_space")
            if result is not None:
                names = result.CHANNEL_NAMES
            # If requested attribute is a channel name, return the attribute from the Space instance.
            if name in names:
                return getattr(result, name)

    def __setattr__(self, name, value):
        """Set attribute."""

        try:
            # See if we need to set the space specific channel attributes.
            if name in self._space.CHANNEL_NAMES:
                setattr(self._space, name, value)
                return
        except AttributeError:
            pass
        # Set all attributes on the Color class.
        super().__setattr__(name, value)



"""HSV class."""
from ..spaces import Space, RE_DEFAULT_MATCH, Angle, Percent, GamutBound, Cylindrical
from .srgb import SRGB
from .hsl import HSL
from .. import util
import re


def hsv_to_hsl(hsv):
    """
    HSV to HSL.

    https://en.wikipedia.org/wiki/HSL_and_HSV#Interconversion
    """

    h, s, v = hsv
    s /= 100.0
    v /= 100.0
    l = v * (1.0 - s / 2.0)
    s = 0.0 if (l == 0.0 or l == 1.0) else ((v - l) / min(l, 1.0 - l)) * 100

    if s == 0:
        h = util.NaN

    return [
        util.constrain_hue(h),
        s,
        l * 100
    ]


def hsl_to_hsv(hsl):
    """
    HSL to HSV.

    https://en.wikipedia.org/wiki/HSL_and_HSV#Interconversion
    """

    h, s, l = hsl
    s /= 100.0
    l /= 100.0

    v = l + s * min(l, 1.0 - l)
    s = 0.0 if (v == 0.0) else 2 * (1.0 - l / v)

    if s == 0:
        h = util.NaN

    return [util.constrain_hue(h), s * 100.0, v * 100.0]


class HSV(Cylindrical, Space):
    """HSL class."""

    SPACE = "hsv"
    CHANNEL_NAMES = ("hue", "saturation", "value", "alpha")
    DEFAULT_MATCH = re.compile(RE_DEFAULT_MATCH.format(color_space=SPACE))
    GAMUT_CHECK = "hsl"
    WHITE = "D65"

    RANGE = (
        GamutBound([Angle(0.0), Angle(360.0)]),
        GamutBound([Percent(0.0), Percent(100.0)]),
        GamutBound([Percent(0.0), Percent(100.0)])
    )

    @property
    def hue(self):
        """Hue channel."""

        return self._coords[0]

    @hue.setter
    def hue(self, value):
        """Shift the hue."""

        self._coords[0] = self._handle_input(value)

    @property
    def saturation(self):
        """Saturation channel."""

        return self._coords[1]

    @saturation.setter
    def saturation(self, value):
        """Saturate or unsaturate the color by the given factor."""

        self._coords[1] = self._handle_input(value)

    @property
    def value(self):
        """Value channel."""

        return self._coords[2]

    @value.setter
    def value(self, value):
        """Set value channel."""

        self._coords[2] = self._handle_input(value)

    @classmethod
    def null_adjust(cls, coords, alpha):
        """On color update."""

        if coords[1] == 0:
            coords[0] = util.NaN
        return coords, alpha

    @classmethod
    def _to_xyz(cls, parent, hsv):
        """To XYZ."""

        return SRGB._to_xyz(parent, cls._to_srgb(parent, hsv))

    @classmethod
    def _from_xyz(cls, parent, xyz):
        """From XYZ."""

        return cls._from_srgb(parent, SRGB._from_xyz(parent, xyz))

    @classmethod
    def _to_hsl(cls, parent, hsv):
        """To HSL."""

        return hsv_to_hsl(hsv)

    @classmethod
    def _from_hsl(cls, parent, hsl):
        """From HSL."""

        return hsl_to_hsv(hsl)

    @classmethod
    def _to_srgb(cls, parent, hsv):
        """To sRGB."""

        return HSL._to_srgb(parent, cls._to_hsl(parent, hsv))

    @classmethod
    def _from_srgb(cls, parent, rgb):
        """From sRGB."""

        return cls._from_hsl(parent, HSL._from_srgb(parent, rgb))



"""Lab D65 class."""
from ..spaces import RE_DEFAULT_MATCH
from .xyz import XYZ
from .lab import LabBase, lab_to_xyz, xyz_to_lab
import re


class LabD65(LabBase):
    """Lab D65 class."""

    SPACE = "lab-d65"
    DEFAULT_MATCH = re.compile(RE_DEFAULT_MATCH.format(color_space=SPACE))
    WHITE = "D65"

    @classmethod
    def _to_xyz(cls, parent, lab):
        """To XYZ."""

        return parent.chromatic_adaptation(cls.WHITE, XYZ.WHITE, lab_to_xyz(lab, cls.white()))

    @classmethod
    def _from_xyz(cls, parent, xyz):
        """From XYZ."""

        return xyz_to_lab(parent.chromatic_adaptation(XYZ.WHITE, cls.WHITE, xyz), cls.white())



"""Color base."""
from abc import ABCMeta
from .. import util
from . import _parse

# Technically this form can handle any number of channels as long as any
# extra are thrown away. We only support 6 currently. If we ever support
# colors with more channels, we can bump this.
RE_DEFAULT_MATCH = r"""(?xi)
color\(\s*
(?:({{color_space}})\s+)?
((?:{percent}|{float})(?:{space}(?:{percent}|{float})){{{{,6}}}}(?:{slash}(?:{percent}|{float}))?)
\s*\)
""".format(
    **_parse.COLOR_PARTS
)

WHITES = {
    "A": [1.09850, 1.00000, 0.35585],
    "B": [0.99072, 1.00000, 0.85223],
    "C": [0.98074, 1.00000, 1.18232],
    "D50": [0.96422, 1.00000, 0.82521],
    "D55": [0.95682, 1.00000, 0.92149],
    "D65": [0.95047, 1.00000, 1.08883],
    "D75": [0.94972, 1.00000, 1.22638],
    "E": [1.00000, 1.00000, 1.00000],
    "F2": [0.99186, 1.00000, 0.67393],
    "F7": [0.95041, 1.00000, 1.08747],
    "F11": [1.00962, 1.00000, 0.64350]
}


class Angle(float):
    """Angle type."""


class Percent(float):
    """Percent type."""


class GamutBound(tuple):
    """Bounded gamut value."""


class GamutUnbound(tuple):
    """Unbounded gamut value."""


class Cylindrical:
    """Cylindrical space."""

    def hue_name(self):
        """Hue channel name."""

        return "hue"


class Space(
    metaclass=ABCMeta
):
    """Base color space object."""

    # Color space name
    SPACE = ""
    # Number of channels
    NUM_COLOR_CHANNELS = 3
    # Channel names
    CHANNEL_NAMES = ("alpha",)
    # For matching the default form of `color(space coords+ / alpha)`.
    # Classes should define this if they want to use the default match.
    DEFAULT_MATCH = ""
    # Match pattern variable for classes to override so we can also
    # maintain the default and other alternatives.
    MATCH = ""
    # Should this color also be checked in a different color space? Only when set to a string (specifying a color space)
    # will the default gamut checking also check the specified space as well as the current.
    #
    # Gamut checking:
    #   The specified color space will be checked first followed by the original. Assuming the parent color space fits,
    #   the original should fit as well, but there are some cases when a parent color space that is slightly out of
    #   gamut, when evaluated with a threshold, may appear to be in gamut enough, but when checking the original color
    #   space, the values can be greatly out of specification (looking at you HSL).
    GAMUT_CHECK = None
    # White point
    WHITE = "D50"

    def __init__(self, color, alpha=None):
        """Initialize."""

        self._alpha = util.NaN
        self._coords = [util.NaN] * self.NUM_COLOR_CHANNELS

        if isinstance(color, Space):
            for index, channel in enumerate(color.coords()):
                self.set(self.CHANNEL_NAMES[index], channel)
            self.alpha = color.alpha
        elif isinstance(color, (list, tuple)):
            if len(color) != self.NUM_COLOR_CHANNELS:  # pragma: no cover
                # Only likely to happen with direct usage internally.
                raise ValueError(
                    "A list of channel values should be at a minimum of {}.".format(self.NUM_COLOR_CHANNELS)
                )
            for index in range(self.NUM_COLOR_CHANNELS):
                self.set(self.CHANNEL_NAMES[index], color[index])
            self.alpha = 1.0 if alpha is None else alpha
        else:  # pragma: no cover
            # Only likely to happen with direct usage internally.
            raise TypeError("Unexpected type '{}' received".format(type(color)))

    def __repr__(self):
        """Representation."""

        gamut = self.RANGE
        values = []
        for i, coord in enumerate(util.no_nan(self.coords())):
            value = util.fmt_float(coord, util.DEF_PREC)
            if isinstance(gamut[i][0], Percent):
                value += '%'
            values.append(value)

        return 'color({} {} / {})'.format(
            self.space(),
            ' '.join(values),
            util.fmt_float(util.no_nan(self.alpha), util.DEF_PREC)
        )

    __str__ = __repr__

    def _handle_input(self, value):
        """Handle numerical input."""

        if not util.is_number(value):
            raise TypeError("Value should be a number not type '{}'".format(type(value)))
        return float(value) if not util.is_nan(value) else value

    def coords(self):
        """Coordinates."""

        return self._coords[:]

    @classmethod
    def space(cls):
        """Get the color space."""

        return cls.SPACE

    @classmethod
    def white(cls):
        """Get the white color for this color space."""

        return WHITES[cls.WHITE]

    @property
    def alpha(self):
        """Alpha channel."""

        return self._alpha

    @alpha.setter
    def alpha(self, value):
        """Adjust alpha."""

        self._alpha = util.clamp(self._handle_input(value), 0.0, 1.0)

    def set(self, name, value):  # noqa: A003
        """Set the given channel."""

        if name not in self.CHANNEL_NAMES:
            raise ValueError("'{}' is an invalid channel name".format(name))

        setattr(self, name, value)
        return self

    def get(self, name):
        """Get the given channel's value."""

        if name not in self.CHANNEL_NAMES:
            raise ValueError("'{}' is an invalid channel name".format(name))
        return getattr(self, name)

    def to_string(
        self, parent, *, alpha=None, precision=None, fit=True, **kwargs
    ):
        """Convert to CSS 'color' string: `color(space coords+ / alpha)`."""

        if precision is None:
            precision = parent.PRECISION

        a = util.no_nan(self.alpha)
        alpha = alpha is not False and (alpha is True or a < 1.0)

        method = None if not isinstance(fit, str) else fit
        coords = util.no_nan(parent.fit(method=method).coords() if fit else self.coords())
        gamut = self.RANGE
        template = "color({} {} / {})" if alpha else "color({} {})"

        values = []
        for i, coord in enumerate(coords):
            value = util.fmt_float(coord, precision)
            if isinstance(gamut[i][0], Percent):
                value += '%'
            values.append(value)

        if alpha:
            return template.format(self.space(), ' '.join(values), util.fmt_float(a, max(precision, util.DEF_PREC)))
        else:
            return template.format(self.space(), ' '.join(values))

    @classmethod
    def null_adjust(cls, coords, alpha):
        """Process coordinates and adjust any channels to null/NaN if required."""

        return coords, alpha

    @classmethod
    def match(cls, string, start=0, fullmatch=True):
        """Match a color by string."""

        m = cls.DEFAULT_MATCH.match(string, start)
        if (
            m is not None and
            (
                (m.group(1) and m.group(1).lower() == cls.space())
            ) and (not fullmatch or m.end(0) == len(string))
        ):

            # Break channels up into a list
            split = _parse.RE_SLASH_SPLIT.split(m.group(2).strip(), maxsplit=1)

            # Get alpha channel
            alpha = _parse.norm_alpha_channel(split[-1]) if len(split) > 1 else 1.0

            # Parse color channels
            channels = []
            for i, c in enumerate(_parse.RE_CHAN_SPLIT.split(split[0]), 0):
                if c and i < cls.NUM_COLOR_CHANNELS:
                    is_percent = isinstance(cls.RANGE[i][0], Percent)
                    if is_percent and not c.endswith('%'):
                        # We have an invalid percentage channel
                        return None, None
                    channels.append(_parse.norm_color_channel(c, not is_percent))

            # Missing channels are filled with zeros
            if len(channels) < cls.NUM_COLOR_CHANNELS:
                diff = cls.NUM_COLOR_CHANNELS - len(channels)
                channels.extend([0.0] * diff)

            # Apply null adjustments (null hues) if applicable
            return cls.null_adjust(channels, alpha), m.end(0)

        return None, None



"""Utilities."""
import copy
import decimal
import math
import numbers
import re
import warnings
from functools import wraps

RE_FLOAT_TRIM = re.compile(r'^(?P<keep>-?\d+)(?P<trash>\.0+|(?P<keep2>\.\d*[1-9])0+)$')
NaN = float('nan')
INF = float('inf')
ACHROMATIC_THRESHOLD = 0.0005
DEF_PREC = 5
DEF_FIT_TOLERANCE = 0.000075
DEF_ALPHA = 1.0
DEF_MIX = 0.5
DEF_HUE_ADJ = "shorter"
DEF_DISTANCE_SPACE = "lab"
DEF_FIT = "lch-chroma"
DEF_DELTA_E = "76"

# Many libraries use 200, but `Colorjs.io` uses 203
# The author explains why 203 was chosen:
#
#   Maximum luminance in PQ is 10,000 cd/m^2
#   Relative XYZ has Y=1 for media white
#   BT.2048 says media white Y=203 at PQ 58
#
# We will currently use 203 for now as the difference is minimal.
# If there were a significant difference, and one clearly gave
# better results, that would make the decision easier, but the
# explanation above seems sufficient for now.
YW = 203

# PQ Constants
# https://en.wikipedia.org/wiki/High-dynamic-range_video#Perceptual_quantizer
M1 = 2610 / 16384
M2 = 2523 / 32
C1 = 3424 / 4096
C2 = 2413 / 128
C3 = 2392 / 128


def pq_st2084_inverse_eotf(values, c1=C1, c2=C2, c3=C3, m1=M1, m2=M2):
    """Perceptual quantizer (SMPTE ST 2084) - inverse EOTF."""

    adjusted = []
    for c in values:
        c = npow(c / 10000, m1)
        r = (c1 + c2 * c) / (1 + c3 * c)
        adjusted.append(npow(r, m2))
    return adjusted


def pq_st2084_eotf(values, c1=C1, c2=C2, c3=C3, m1=M1, m2=M2):
    """Perceptual quantizer (SMPTE ST 2084) - EOTF."""

    im1 = 1 / m1
    im2 = 1 / m2

    adjusted = []
    for c in values:
        c = npow(c, im2)
        r = (c - c1) / (c2 - c3 * c)
        adjusted.append(10000 * npow(r, im1))
    return adjusted


def xyz_d65_to_absxyzd65(xyzd65):
    """XYZ D65 to Absolute XYZ D65."""

    return [max(c * YW, 0) for c in xyzd65]


def absxyzd65_to_xyz_d65(absxyzd65):
    """Absolute XYZ D65 XYZ D65."""

    return [max(c / YW, 0) for c in absxyzd65]


def npow(base, exp):
    """Perform `pow` with a negative number."""

    return math.copysign(abs(base) ** exp, base)


def constrain_hue(hue):
    """Constrain hue to 0 - 360."""

    return hue % 360 if not is_nan(hue) else hue


def is_number(value):
    """Check if value is a number."""

    return isinstance(value, numbers.Number)


def is_nan(value):
    """Check if value is "not a number"."""

    return math.isnan(value)


def no_nan(value):
    """Convert list of numbers or single number to valid numbers."""

    if is_number(value):
        return 0.0 if is_nan(value) else value
    else:
        return [(0.0 if is_nan(x) else x) for x in value]


def cmp_coords(c1, c2):
    """Compare coordinates."""

    if is_number(c1):
        return (math.isnan(c1) and math.isnan(c2)) or c1 == c2
    elif len(c1) != len(c2):
        return False
    else:
        return all(map(lambda a, b: (math.isnan(a) and math.isnan(b)) or a == b, c1, c2))


def dot(a, b):
    """Get dot product of simple numbers, vectors, and 2D matrices and/or numbers."""

    is_a_num = is_number(a)
    is_b_num = is_number(b)
    is_a_vec = not is_a_num and is_number(a[0])
    is_b_vec = not is_b_num and is_number(b[0])
    is_a_mat = not is_a_num and not is_a_vec
    is_b_mat = not is_b_num and not is_b_vec

    if is_a_num or is_b_num:
        # Trying to dot a number with a vector or a matrix, so just multiply
        value = multiply(a, b)
    elif is_a_vec and is_b_vec:
        # Dot product of two vectors
        value = sum([x * y for x, y in zip(a, b)])
    elif is_a_mat and is_b_vec:
        # Dot product of matrix and a vector
        value = [sum([x * y for x, y in zip(row, b)]) for row in a]
    elif is_a_vec and is_b_mat:
        # Dot product of vector and a matrix
        value = [sum([x * y for x, y in zip(a, col)]) for col in zip(*b)]
    else:
        # Dot product of two matrices
        value = [[sum(x * y for x, y in zip(row, col)) for col in zip(*b)] for row in a]

    return value


def multiply(a, b):
    """Multiply simple numbers, vectors, and 2D matrices."""

    is_a_num = is_number(a)
    is_b_num = is_number(b)
    is_a_vec = not is_a_num and is_number(a[0])
    is_b_vec = not is_b_num and is_number(b[0])
    is_a_mat = not is_a_num and not is_a_vec
    is_b_mat = not is_b_num and not is_b_vec

    if is_a_num and is_b_num:
        # Multiply two numbers
        value = a * b
    elif is_a_num and not is_b_num:
        # Multiply a number and vector/matrix
        value = [multiply(a, i) for i in b]
    elif is_b_num and not is_a_num:
        # Multiply a vector/matrix and number
        value = [multiply(i, b) for i in a]
    elif is_a_vec and is_b_vec:
        # Multiply two vectors
        value = [x * y for x, y in zip(a, b)]
    elif is_a_mat and is_b_vec:
        # Multiply matrix and a vector
        value = [[x * y for x, y in zip(row, b)] for row in a]
    elif is_a_vec and is_b_mat:
        # Multiply vector and a matrix
        value = [[x * y for x, y in zip(row, a)] for row in b]
    else:
        # Multiply two matrices
        value = [[x * y for x, y in zip(ra, rb)] for ra, rb in zip(a, b)]

    return value


def divide(a, b):
    """Divide simple numbers, vectors, and 2D matrices."""

    is_a_num = isinstance(a, numbers.Number)
    is_b_num = isinstance(b, numbers.Number)
    is_a_vec = not is_a_num and isinstance(a[0], numbers.Number)
    is_b_vec = not is_b_num and isinstance(b[0], numbers.Number)
    is_a_mat = not is_a_num and not is_a_vec
    is_b_mat = not is_b_num and not is_b_vec

    if is_a_num and is_b_num:
        # Divide two numbers
        value = a / b
    elif is_a_num and not is_b_num:
        # Divide a number and vector/matrix
        value = [divide(a, i) for i in b]
    elif is_b_num and not is_a_num:
        # Divide a vector/matrix and number
        value = [divide(i, b) for i in a]
    elif is_a_vec and is_b_vec:
        # Divide two vectors
        value = [x / y for x, y in zip(a, b)]
    elif is_a_mat and is_b_vec:
        # Divide matrix and a vector
        value = [[x / y for x, y in zip(row, b)] for row in a]
    elif is_a_vec and is_b_mat:
        # Divide vector and a matrix
        value = [[x / y for x, y in zip(row, a)] for row in b]
    else:
        # Divide two matrices
        value = [[x / y for x, y in zip(ra, rb)] for ra, rb in zip(a, b)]

    return value


def diag(v, k=0):
    """Create a diagonal matrix from a vector or return a vector of the diagonal of a matrix."""

    is_vector = isinstance(v[0], numbers.Number)
    size = len(v)
    d = []

    if is_vector:
        # Create a diagonal matrix with the provided values
        for i, value in enumerate(v):
            # Check that the matrix is square, we .cannot invert the matrix if it is not
            d.append([0] * i + [value] + [0] * (size - i - 1))
    else:  # pragma: no cover
        for r in v:
            if len(r) != size:
                raise ValueError('Matrix must be a n x n matrix')
            if 0 <= k < size:
                d.append(r[k])
            k += 1
    return d


def inv(matrix):
    """
    Invert the matrix.

    Derived from https://github.com/ThomIves/MatrixInverse.

    While not as performant as using `numpy`, we are often caching any
    inversion we are doing, so this keeps us from having to require all
    of `numpy` for the few hits to this we do.

    This is free and unencumbered software released into the public domain.

    Anyone is free to copy, modify, publish, use, compile, sell, or
    distribute this software, either in source code form or as a compiled
    binary, for any purpose, commercial or non-commercial, and by any
    means.

    In jurisdictions that recognize copyright laws, the author or authors
    of this software dedicate any and all copyright interest in the
    software to the public domain. We make this dedication for the benefit
    of the public at large and to the detriment of our heirs and
    successors. We intend this dedication to be an overt act of
    relinquishment in perpetuity of all present and future rights to this
    software under copyright law.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
    MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
    IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR
    OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
    ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
    OTHER DEALINGS IN THE SOFTWARE.

    For more information, please refer to <http://unlicense.org/>
    """

    size = len(matrix)
    indices = list(range(size))
    m = copy.deepcopy(matrix)

    # Ensure we have a square matrix
    for r in m:
        if len(r) != size:  # pragma: no cover
            raise ValueError('Matrix must be a n x n matrix')

    # Create an identity matrix of the same size as our provided matrix
    im = diag([1] * size)

    # Iterating through each row, we will scale each row by it's "focus diagonal".
    # Then using the scaled row, we will adjust the other rows.
    # ```
    # [[fd, 0,  0 ]
    #  [0,  fd, 0 ]
    #  [0,  0,  fd]]
    # ```
    for fd in indices:
        # We will divide each value in the row by the "focus diagonal" value.
        # If the we have a zero for the given `fd` value, we cannot invert.
        denom = m[fd][fd]
        if denom == 0:  # pragma: no cover
            raise ValueError('Matrix is not invertable')

        # We are converting the matrix to the identity and vice versa,
        # So scale the diagonal such that it will now equal 1.
        # Additionally, the same operations will be applied to the identity matrix
        # and will turn it into `m ** -1` (what we are looking for)
        fd_scalar = 1.0 / denom
        for j in indices:
            m[fd][j] *= fd_scalar
            im[fd][j] *= fd_scalar

        # Now, using the value found at the index `fd` in the remaining rows (excluding `row[fd]`),
        # Where `cr` is the current row under evaluation, subtract `row[cr][fd] * row[fd] from row[cr]`.
        for cr in indices[0:fd] + indices[fd + 1:]:
            # The scalar for the current row
            cr_scalar = m[cr][fd]

            # Scale each item in the `row[fd]` and subtract it from the current row `row[cr]`
            for j in indices:
                m[cr][j] -= cr_scalar * m[fd][j]
                im[cr][j] -= cr_scalar * im[fd][j]

    # The identify matrix is now the inverse matrix and vice versa.
    return im


def cbrt(x):
    """Cube root."""

    if 0 <= x:
        return x ** (1.0 / 3.0)
    return -(-x) ** (1.0 / 3.0)


def clamp(value, mn=None, mx=None):
    """Clamp the value to the the given minimum and maximum."""

    if mn is None and mx is None:
        return value
    elif mn is None:
        return min(value, mx)
    elif mx is None:
        return max(value, mn)
    else:
        return max(min(value, mx), mn)


def adjust_precision(f, p):
    """Adjust precision and scale."""

    with decimal.localcontext() as ctx:
        if p > 0:
            # Set precision
            ctx.prec = p
        ctx.rounding = decimal.ROUND_HALF_UP

        if p == -1:
            # Full precision
            value = decimal.Decimal(f)
        elif p == 0:
            # Just round to integer
            value = decimal.Decimal(round_half_up(f))
        else:
            # Round to precision
            value = (decimal.Decimal(f) * decimal.Decimal('1.0'))
            exp = value.as_tuple().exponent
            if exp < 0 and abs(value.as_tuple().exponent) > p:
                value = value.quantize(decimal.Decimal(10) ** -p)

        if value.is_zero():
            value = abs(value)

        return float(value)


def fmt_float(f, p=0):
    """
    Set float precision and trim precision zeros.

    0: Round to whole integer
    -1: Full precision
    <positive number>: precision level
    """

    value = adjust_precision(f, p)
    string = ('{{:{}f}}'.format('.53' if p == -1 else '.' + str(p))).format(value)
    m = RE_FLOAT_TRIM.match(string)
    if m:
        string = m.group('keep')
        if m.group('keep2'):
            string += m.group('keep2')
    return string


def round_half_up(n, scale=0):
    """Round half up."""

    if scale == -1:
        return n

    mult = 10 ** scale
    return math.floor(n * mult + 0.5) / mult


def deprecated(message, stacklevel=2):  # pragma: no cover
    """
    Raise a `DeprecationWarning` when wrapped function/method is called.

    Borrowed from https://stackoverflow.com/a/48632082/866026
    """

    def _decorator(func):
        @wraps(func)
        def _func(*args, **kwargs):
            warnings.warn(
                "'{}' is deprecated. {}".format(func.__name__, message),
                category=DeprecationWarning,
                stacklevel=stacklevel
            )
            return func(*args, **kwargs)
        return _func
    return _decorator



"""
Delta E ITP.

https://kb.portrait.com/help/ictcp-color-difference-metric
"""
import math


def distance(color1, color2, scalar=720, **kwargs):
    """Delta E ITP color distance formula."""

    i1, t1, p1 = color1.convert('ictcp').coords()
    i2, t2, p2 = color2.convert('ictcp').coords()

    # Equation (1)
    return scalar * math.sqrt((i2 - i1) ** 2 + 0.25 * (t2 - t1) ** 2 + (p2 - p1) ** 2)



"""Fit by compressing chroma in Lch."""

EPSILON = 0.001


def fit(color):
    """
    Gamut mapping via chroma Lch.

    Algorithm originally came from https://colorjs.io/docs/gamut-mapping.html.
    Some things have been optimized and fixed though to better perform as intended.

    The idea is to hold hue and lightness constant and decrease lightness until
    color comes under gamut.

    We'll use a binary search and at after each stage, we will clip the color
    and compare the distance of the two colors (clipped and current color via binary search).
    If the distance is less than two, we can return the clipped color.

    ---
    Original Authors: Lea Verou, Chris Lilley
    License: MIT (As noted in https://github.com/LeaVerou/color.js/blob/master/package.json)
    """

    space = color.space()

    # If flooring chroma doesn't work, just clip the floored color
    # because there is no optimal compression.
    floor = color.clone().set('lch.chroma', 0)
    if not floor.in_gamut():
        return floor.fit(method="clip").coords()

    # If we are already below the JND, just clip as we will gain no
    # noticeable difference moving forward.
    clipped = color.fit(method="clip")
    if color.delta_e(clipped, method="2000") < 2.3:
        return clipped.coords()

    # Convert to CIELCH and set our boundaries
    mapcolor = color.convert("lch")
    low = 0.0
    high = mapcolor.chroma

    # Adjust chroma (using binary search).
    # This helps preserve the other attributes of the color.
    # Each time we compare the compressed color to it's clipped form
    # to see how close we are. A delta less than 2 is our target.
    while (high - low) > EPSILON:
        delta = mapcolor.delta_e(
            mapcolor.fit(space, method="clip"),
            method="2000"
        )

        if (delta - 2) < EPSILON:
            low = mapcolor.chroma
        else:
            high = mapcolor.chroma

        mapcolor.chroma = (high + low) * 0.5

    # Update and clip off noise
    return color.update(mapcolor).fit(space, method="clip", in_place=True).coords()



"""Lch class."""
from ..spaces import Space, RE_DEFAULT_MATCH, GamutUnbound, Cylindrical, Angle, Percent
from .lab import Lab
from .. import util
import re
import math

ACHROMATIC_THRESHOLD = 0.00000000001


def lab_to_lch(lab):
    """Lab to Lch."""

    l, a, b = lab

    c = math.sqrt(a ** 2 + b ** 2)
    h = math.degrees(math.atan2(b, a))

    # Achromatic colors will often get extremely close, but not quite hit zero.
    # Essentially, we want to discard noise through rounding and such.
    if c < ACHROMATIC_THRESHOLD:
        h = util.NaN

    test = [l, c, util.constrain_hue(h)]
    return test


def lch_to_lab(lch):
    """Lch to Lab."""

    l, c, h = lch
    h = util.no_nan(h)

    # If, for whatever reason (mainly direct user input),
    # if chroma is less than zero, clamp to zero.
    if c < 0.0:
        c = 0.0

    return (
        l,
        c * math.cos(math.radians(h)),
        c * math.sin(math.radians(h))
    )


class LchBase(Cylindrical, Space):
    """Lch class."""

    CHANNEL_NAMES = ("lightness", "chroma", "hue", "alpha")

    RANGE = (
        # I think chroma, specifically should be clamped.
        # Some libraries don't to prevent rounding issues. We should only get
        # negative chroma via direct user input, but when translating to
        # Lab, this will be corrected.
        GamutUnbound([Percent(0.0), Percent(100.0)]),
        GamutUnbound([0.0, 100.0]),
        GamutUnbound([Angle(0.0), Angle(360.0)]),
    )

    @property
    def lightness(self):
        """Lightness."""

        return self._coords[0]

    @lightness.setter
    def lightness(self, value):
        """Get true luminance."""

        self._coords[0] = self._handle_input(value)

    @property
    def chroma(self):
        """Chroma."""

        return self._coords[1]

    @chroma.setter
    def chroma(self, value):
        """chroma."""

        self._coords[1] = self._handle_input(value)

    @property
    def hue(self):
        """Hue."""

        return self._coords[2]

    @hue.setter
    def hue(self, value):
        """Shift the hue."""

        self._coords[2] = self._handle_input(value)

    @classmethod
    def null_adjust(cls, coords, alpha):
        """On color update."""

        if coords[1] < ACHROMATIC_THRESHOLD:
            coords[2] = util.NaN
        return coords, alpha


class Lch(LchBase):
    """Lch class."""

    SPACE = "lch"
    DEFAULT_MATCH = re.compile(RE_DEFAULT_MATCH.format(color_space=SPACE))
    WHITE = "D50"

    @classmethod
    def _to_lab(cls, parent, lch):
        """To Lab."""

        return lch_to_lab(lch)

    @classmethod
    def _from_lab(cls, parent, lab):
        """To Lab."""

        return lab_to_lch(lab)

    @classmethod
    def _to_xyz(cls, parent, lch):
        """To XYZ."""

        return Lab._to_xyz(parent, cls._to_lab(parent, lch))

    @classmethod
    def _from_xyz(cls, parent, xyz):
        """From XYZ."""

        return cls._from_lab(parent, Lab._from_xyz(parent, xyz))



"""Rec 2020 color class."""
from ..spaces import RE_DEFAULT_MATCH
from .srgb import SRGB
from .xyz import XYZ
from .. import util
import re
import math

ALPHA = 1.09929682680944
BETA = 0.018053968510807
BETA45 = 0.018053968510807 * 4.5


def lin_2020(rgb):
    """
    Convert an array of rec-2020 RGB values in the range 0.0 - 1.0 to linear light (un-corrected) form.

    https://en.wikipedia.org/wiki/Rec._2020#Transfer_characteristics
    """

    result = []
    for i in rgb:
        # Mirror linear nature of algorithm on the negative axis
        abs_i = abs(i)
        if abs_i < BETA45:
            result.append(i / 4.5)
        else:
            result.append(math.copysign(((abs_i + ALPHA - 1) / ALPHA) ** (1 / 0.45), i))
    return result


def gam_2020(rgb):
    """
    Convert an array of linear-light rec-2020 RGB  in the range 0.0-1.0 to gamma corrected form.

    https://en.wikipedia.org/wiki/Rec._2020#Transfer_characteristics
    """

    result = []
    for i in rgb:
        # Mirror linear nature of algorithm on the negative axis
        abs_i = abs(i)
        if abs_i < BETA:
            result.append(4.5 * i)
        else:
            result.append(math.copysign(ALPHA * (abs_i ** 0.45) - (ALPHA - 1), i))
    return result


def lin_2020_to_xyz(rgb):
    """
    Convert an array of linear-light rec-2020 values to CIE XYZ using  D65.

    (no chromatic adaptation)
    http://www.brucelindbloom.com/index.html?Eqn_RGB_XYZ_Matrix.html
    """

    m = [
        [6.3701019141110093e-01, 1.4461502739696933e-01, 1.6884478119192992e-01],
        [2.6272171736164052e-01, 6.7798927550226207e-01, 5.9289007136097520e-02],
        [4.9945154055471928e-17, 2.8072328847646915e-02, 1.0607576711523534e+00]
    ]

    return util.dot(m, rgb)


def xyz_to_lin_2020(xyz):
    """Convert XYZ to linear-light rec-2020."""

    m = [
        [1.7165106697619734, -0.3556416699867159, -0.2533455418219072],
        [-0.6666930011826241, 1.6165022083469103, 0.015768750389995],
        [0.017643638767459, -0.0427797816690446, 0.9423050727200183]
    ]

    return util.dot(m, xyz)


class Rec2020(SRGB):
    """Rec 2020 class."""

    SPACE = "rec2020"
    DEFAULT_MATCH = re.compile(RE_DEFAULT_MATCH.format(color_space=SPACE))
    WHITE = "D65"

    @classmethod
    def _to_xyz(cls, parent, rgb):
        """To XYZ."""

        return parent.chromatic_adaptation(cls.WHITE, XYZ.WHITE, lin_2020_to_xyz(lin_2020(rgb)))

    @classmethod
    def _from_xyz(cls, parent, xyz):
        """From XYZ."""

        return gam_2020(xyz_to_lin_2020(parent.chromatic_adaptation(XYZ.WHITE, cls.WHITE, xyz)))



"""Delta E CMC."""
import math
from ... import util


def distance(color1, color2, l=2, c=1):
    """
    Delta E CMC.

    http://www.brucelindbloom.com/index.html?Eqn_DeltaE_CMC.html
    """

    l1, a1, b1 = util.no_nan(color1.convert("lab").coords())
    l2, a2, b2 = util.no_nan(color2.convert("lab").coords())

    # Equation (3)
    c1 = math.sqrt(a1 ** 2 + b1 ** 2)

    # Equation (4)
    c2 = math.sqrt(a2 ** 2 + b2 ** 2)

    # Equation (2)
    dc = c1 - c2

    # Equation (6)
    dl = l1 - l2

    # Equation (7)
    da = a1 - a2

    # Equation (8)
    db = b1 - b2

    # Equation (5)
    # We never reference `dh` until the very end, and when we do, we square it
    # before using it, so we don't need the square root as described in the
    # algorithm. Instead we can just leave the result as is.
    dh = da ** 2 + db ** 2 - dc ** 2

    # Equation (9)
    if l1 < 16:
        sl = 0.511
    else:
        sl = (0.040975 * l1) / (1 + 0.01765 * l1)

    # Equation (10)
    sc = ((0.0638 * c1) / (1 + 0.0131 * c1)) + 0.638

    # Equation (14)
    h = math.degrees(math.atan2(b1, a1))

    # Equation (15)
    if h >= 0:
        h1 = h
    else:
        h1 = h + 360

    # Equation (12)
    if 164 <= h1 <= 345:
        t = 0.56 + abs(0.2 * math.cos(math.radians(h1 + 168)))
    else:
        t = 0.36 + abs(0.4 * math.cos(math.radians(h1 + 35)))

    # Equation (13)
    c1_4 = c1 ** 4
    f = math.sqrt(c1_4 / (c1_4 + 1900))

    # Equation (11)
    sh = sc * (f * t + 1 - f)

    # Equation (1)
    return math.sqrt(
        (dl / (l * sl)) ** 2 +
        (dc / (c * sc)) ** 2 +
        # Square root just the denominator as `dh` is already squared.
        dh / (sh ** 2)
    )



"""Display-p3 color class."""
from ..spaces import RE_DEFAULT_MATCH
from .srgb import SRGB, lin_srgb, gam_srgb
from .xyz import XYZ
from .. import util
import re


def lin_p3_to_xyz(rgb):
    """
    Convert an array of linear-light image-p3 values to CIE XYZ using  D65 (no chromatic adaptation).

    http://www.brucelindbloom.com/index.html?Eqn_RGB_XYZ_Matrix.html
    """

    m = [
        [4.8663264999999994e-01, 2.6566316250000005e-01, 1.9817418749999988e-01],
        [2.2900359999999997e-01, 6.9172672500000010e-01, 7.9269674999999956e-02],
        [-3.9725792100320233e-17, 4.5112612500000052e-02, 1.0437173874999994e+00]
    ]

    # 0 was computed as -3.972075516933488e-17
    return util.dot(m, rgb)


def xyz_to_lin_p3(xyz):
    """Convert XYZ to linear-light P3."""

    m = [
        [2.493180755328967, -0.9312655254971399, -0.4026597237588819],
        [-0.8295031158210786, 1.7626941211197922, 0.0236250887417396],
        [0.0358536257800717, -0.0761889547826522, 0.9570926215180221]
    ]

    return util.dot(m, xyz)


def lin_p3(rgb):
    """Convert an array of image-p3 RGB values in the range 0.0 - 1.0 to linear light (un-corrected) form."""

    return lin_srgb(rgb)  # same as sRGB


def gam_p3(rgb):
    """Convert an array of linear-light image-p3 RGB  in the range 0.0-1.0 to gamma corrected form."""

    return gam_srgb(rgb)  # same as sRGB


class DisplayP3(SRGB):
    """Display-p3 class."""

    SPACE = "display-p3"
    DEFAULT_MATCH = re.compile(RE_DEFAULT_MATCH.format(color_space=SPACE))
    WHITE = "D65"

    @classmethod
    def _to_xyz(cls, parent, rgb):
        """To XYZ."""

        return parent.chromatic_adaptation(cls.WHITE, XYZ.WHITE, lin_p3_to_xyz(lin_p3(rgb)))

    @classmethod
    def _from_xyz(cls, parent, xyz):
        """From XYZ."""

        return gam_p3(xyz_to_lin_p3(parent.chromatic_adaptation(XYZ.WHITE, cls.WHITE, xyz)))



"""
Compositing and RGB blend modes.

https://www.w3.org/TR/compositing/
"""
from collections.abc import Sequence
from . import porter_duff
from . import blend_modes
from ... import util
from ...spaces import GamutBound


def clip_channel(coord, gamut):
    """Clipping channel."""

    a, b = gamut
    is_bound = isinstance(gamut, GamutBound)

    # These parameters are unbounded
    if not is_bound:  # pragma: no cover
        # Will not execute unless we have a space that defines some coordinates
        # as bound and others as not. We do not currently have such spaces.
        a = None
        b = None

    # Fit value in bounds.
    return util.clamp(coord, a, b)


def compose(color1, color2, blend, operator, non_seperable):
    """Blend colors using the specified blend mode."""

    # Get the color coordinates
    csa = util.no_nan(color1.alpha)
    cba = util.no_nan(color2.alpha)
    coords1 = util.no_nan(color1.coords())
    coords2 = util.no_nan(color2.coords())

    # Setup blend mode.
    if blend is None:
        blend = 'normal'
    if blend is not False:
        blend = blend.lower()
        blender = blend_modes.get_blender(blend)
    else:
        blender = None

    # Setup compositing
    if operator is None:
        operator = 'source-over'
    if operator is not False:
        compositor = porter_duff.compositor(operator)(cba, csa)
        cra = compositor.ao()
    else:
        cra = csa
        compositor = None

    # Perform compositing
    gamut = color1._space.RANGE
    coords = []
    if not non_seperable:
        # Blend each channel. Afterward, clip and apply alpha compositing.
        i = 0
        for cb, cs in zip(coords2, coords1):
            cr = (1 - cba) * cs + cba * blender(cb, cs) if blender is not None else cs
            cr = clip_channel(cr, gamut[i])
            coords.append(compositor.co(cb, cr) if compositor is not None else cr)
            i += 1
    else:
        # Convert to a hue, saturation, luminosity space and apply the requested blending.
        # Afterwards, clip and apply alpha compositing.
        i = 0
        blended = blender(coords2, coords1) if blender is not None else coords1
        for cb, cr in zip(coords2, blended):
            cr = (1 - cba) * cr + cba * cr if blender is not None else cr
            cr = clip_channel(cr, gamut[i])
            coords.append(compositor.co(cb, cr) if compositor is not None else cr)
            i += 1

    return color1.update(color1.space(), coords, cra)


class Compose:
    """Handle compositing."""

    def compose(self, backdrop, *, blend=None, operator=None, space=None, out_space=None, in_place=False):
        """Blend colors using the specified blend mode."""

        backdrop = self._handle_color_input(backdrop, sequence=True)

        # If we are doing non-separable, we are converting to a special space that
        # can only be done from sRGB, so we have to force sRGB anyway.
        non_seperable = blend_modes.is_non_seperable(blend)
        if non_seperable:
            space = 'srgb'
        space = 'srgb' if space is None else space.lower()
        outspace = self.space() if out_space is None else out_space.lower()

        if not isinstance(backdrop, Sequence):
            backdrop = [backdrop]

        if len(backdrop) == 0:
            return self.convert(outspace)

        if len(backdrop) > 1:
            dest = backdrop[-1].convert(space, fit=True)
            for x in range(len(backdrop) - 2, -1, -1):
                src = backdrop[x].convert(space, fit=True)
                dest = compose(src, dest, blend, operator, non_seperable)
        else:
            dest = backdrop[0].convert(space, fit=True)

        src = self.convert(space, fit=True)
        dest = compose(src, dest, blend, operator, non_seperable)

        return self.mutate(dest.convert(outspace)) if in_place else dest.convert(outspace)

    @util.deprecated("'overlay' is deprecated, 'compose' should be used instead.")
    def overlay(self, backdrop, *, space=None, in_place=False):
        """Redirect to compose."""

        if space is None:
            space = self.space()
        return self.compose(backdrop, space=space, out_space=None, in_place=in_place)



"""A98 RGB color class."""
from ..spaces import RE_DEFAULT_MATCH
from .srgb import SRGB
from .xyz import XYZ
from .. import util
import re


def lin_a98rgb_to_xyz(rgb):
    """
    Convert an array of linear-light a98-rgb values to CIE XYZ using D50.D65.

    (so no chromatic adaptation needed afterwards)
    http://www.brucelindbloom.com/index.html?Eqn_RGB_XYZ_Matrix.html
    which has greater numerical precision than section 4.3.5.3 of
    https://www.adobe.com/digitalimag/pdfs/AdobeRGB1998.pdf
    """

    m = [
        [0.5767308871981476, 0.1855539507112141, 0.1881851620906385],
        [0.2973768637115448, 0.6273490714522, 0.0752740648362554],
        [0.0270342603374131, 0.0706872193185578, 0.9911085203440293]
    ]

    return util.dot(m, rgb)


def xyz_to_lin_a98rgb(xyz):
    """Convert XYZ to linear-light a98-rgb."""

    m = [
        [2.04136897926008, -0.5649463871751959, -0.3446943843778484],
        [-0.9692660305051867, 1.8760108454466937, 0.0415560175303498],
        [0.0134473872161703, -0.1183897423541256, 1.0154095719504166]
    ]

    return util.dot(m, xyz)


def lin_a98rgb(rgb):
    """Convert an array of a98-rgb values in the range 0.0 - 1.0 to linear light (un-corrected) form."""

    return [util.npow(val, 563 / 256) for val in rgb]


def gam_a98rgb(rgb):
    """Convert an array of linear-light a98-rgb  in the range 0.0-1.0 to gamma corrected form."""

    return [util.npow(val, 256 / 563) for val in rgb]


class A98RGB(SRGB):
    """A98 RGB class."""

    SPACE = "a98-rgb"
    DEFAULT_MATCH = re.compile(RE_DEFAULT_MATCH.format(color_space=SPACE))
    WHITE = "D65"

    @classmethod
    def _to_xyz(cls, parent, rgb):
        """To XYZ."""

        return parent.chromatic_adaptation(cls.WHITE, XYZ.WHITE, lin_a98rgb_to_xyz(lin_a98rgb(rgb)))

    @classmethod
    def _from_xyz(cls, parent, xyz):
        """From XYZ."""

        return gam_a98rgb(xyz_to_lin_a98rgb(parent.chromatic_adaptation(XYZ.WHITE, cls.WHITE, xyz)))



"""Distance and Delta E."""
from . import distance_euclidean
from . import delta_e_76  # noqa: F401
from . import delta_e_94  # noqa: F401
from . import delta_e_cmc  # noqa: F401
from . import delta_e_2000  # noqa: F401
from . import delta_e_itp  # noqa: F401


class Distance:
    """Distance."""

    def delta_e(self, color, *, method=None, **kwargs):
        """Delta E distance."""

        color = self._handle_color_input(color)
        if method is None:
            method = self.DELTA_E

        algorithm = method.lower()

        try:
            de = globals()['delta_e_{}'.format(algorithm.replace('-', '_'))]
        except KeyError:
            raise ValueError("'{}' is not currently a supported distancing algorithm.".format(algorithm))

        return de.distance(self, color, **kwargs)

    def distance(self, color, *, space="lab"):
        """Delta."""

        color = self._handle_color_input(color)
        return distance_euclidean.distance(self, color, space=space)



"""SRGB color class."""
from ..spaces import RE_DEFAULT_MATCH, Space, GamutBound
from .xyz import XYZ
from .. import util
import re
import math


def lin_srgb_to_xyz(rgb):
    """
    Convert an array of linear-light sRGB values to CIE XYZ using sRGB's own white.

    D65 (no chromatic adaptation)
    """

    m = [
        [0.4124564390896923, 0.357576077643909, 0.180437483266399],
        [0.2126728514056226, 0.715152155287818, 0.0721749933065596],
        [0.0193338955823293, 0.119192025881303, 0.950304078536368]
    ]

    return util.dot(m, rgb)


def xyz_to_lin_srgb(xyz):
    """Convert XYZ to linear-light sRGB."""

    m = [
        [3.2404541621141045, -1.5371385127977162, -0.498531409556016],
        [-0.969266030505187, 1.8760108454466944, 0.0415560175303498],
        [0.0556434309591147, -0.2040259135167538, 1.057225188223179]
    ]

    return util.dot(m, xyz)


def lin_srgb(rgb):
    """
    Convert an array of sRGB values in the range 0.0 - 1.0 to linear light (un-corrected) form.

    https://en.wikipedia.org/wiki/SRGB
    """

    result = []
    for i in rgb:
        # Mirror linear nature of algorithm on the negative axis
        abs_i = abs(i)
        if abs_i < 0.04045:
            result.append(i / 12.92)
        else:
            result.append(math.copysign(((abs_i + 0.055) / 1.055) ** 2.4, i))
    return result


def gam_srgb(rgb):
    """
    Convert an array of linear-light sRGB values in the range 0.0-1.0 to gamma corrected form.

    https://en.wikipedia.org/wiki/SRGB
    """

    result = []
    for i in rgb:
        # Mirror linear nature of algorithm on the negative axis
        abs_i = abs(i)
        if abs_i > 0.0031308:
            result.append(math.copysign(1.055 * (abs_i ** (1 / 2.4)) - 0.055, i))
        else:
            result.append(12.92 * i)
    return result


class SRGB(Space):
    """SRGB class."""

    SPACE = "srgb"
    # In addition to the current gamut, check HSL as it is much more sensitive to small
    # gamut changes. This is mainly for a better user experience. Colors will still be
    # mapped/clipped in the current space, unless specified otherwise.
    GAMUT_CHECK = "hsl"
    DEFAULT_MATCH = re.compile(RE_DEFAULT_MATCH.format(color_space=SPACE))
    CHANNEL_NAMES = ("red", "green", "blue", "alpha")
    WHITE = "D65"

    RANGE = (
        GamutBound([0.0, 1.0]),
        GamutBound([0.0, 1.0]),
        GamutBound([0.0, 1.0])
    )

    @property
    def red(self):
        """Adjust red."""

        return self._coords[0]

    @red.setter
    def red(self, value):
        """Adjust red."""

        self._coords[0] = self._handle_input(value)

    @property
    def green(self):
        """Adjust green."""

        return self._coords[1]

    @green.setter
    def green(self, value):
        """Adjust green."""

        self._coords[1] = self._handle_input(value)

    @property
    def blue(self):
        """Adjust blue."""

        return self._coords[2]

    @blue.setter
    def blue(self, value):
        """Adjust blue."""

        self._coords[2] = self._handle_input(value)

    @classmethod
    def _to_xyz(cls, parent, rgb):
        """SRGB to XYZ."""

        return parent.chromatic_adaptation(cls.WHITE, XYZ.WHITE, lin_srgb_to_xyz(lin_srgb(rgb)))

    @classmethod
    def _from_xyz(cls, parent, xyz):
        """XYZ to SRGB."""

        return gam_srgb(xyz_to_lin_srgb(parent.chromatic_adaptation(XYZ.WHITE, cls.WHITE, xyz)))



"""LCH class."""
from ..spaces import Space, RE_DEFAULT_MATCH, GamutUnbound, Cylindrical, Angle
from .oklab import Oklab
from .. import util
import re
import math

ACHROMATIC_THRESHOLD = 0.0002


def oklab_to_oklch(oklab):
    """Oklab to Oklch."""

    l, a, b = oklab

    c = math.sqrt(a ** 2 + b ** 2)
    h = math.degrees(math.atan2(b, a))

    # Achromatic colors will often get extremely close, but not quite hit zero.
    # Essentially, we want to discard noise through rounding and such.
    if c < ACHROMATIC_THRESHOLD:
        h = util.NaN

    return [l, c, util.constrain_hue(h)]


def oklch_to_oklab(oklch):
    """Oklch to Oklab."""

    l, c, h = oklch
    h = util.no_nan(h)

    # If, for whatever reason (mainly direct user input),
    # if chroma is less than zero, clamp to zero.
    if c < 0.0:
        c = 0.0

    return (
        l,
        c * math.cos(math.radians(h)),
        c * math.sin(math.radians(h))
    )


class Oklch(Cylindrical, Space):
    """Oklch class."""

    SPACE = "oklch"
    CHANNEL_NAMES = ("lightness", "chroma", "hue", "alpha")
    DEFAULT_MATCH = re.compile(RE_DEFAULT_MATCH.format(color_space=SPACE))
    WHITE = "D65"

    RANGE = (
        GamutUnbound([0.0, 1.0]),
        GamutUnbound([0.0, 1.0]),
        GamutUnbound([Angle(0.0), Angle(360.0)]),
    )

    @property
    def lightness(self):
        """Lightness."""

        return self._coords[0]

    @lightness.setter
    def lightness(self, value):
        """Get true luminance."""

        self._coords[0] = self._handle_input(value)

    @property
    def chroma(self):
        """Chroma."""

        return self._coords[1]

    @chroma.setter
    def chroma(self, value):
        """chroma."""

        self._coords[1] = self._handle_input(value)

    @property
    def hue(self):
        """Hue."""

        return self._coords[2]

    @hue.setter
    def hue(self, value):
        """Shift the hue."""

        self._coords[2] = self._handle_input(value)

    @classmethod
    def null_adjust(cls, coords, alpha):
        """On color update."""

        if coords[1] < ACHROMATIC_THRESHOLD:
            coords[2] = util.NaN
        return coords, alpha

    @classmethod
    def _to_oklab(cls, parent, oklch):
        """To Lab."""

        return oklch_to_oklab(oklch)

    @classmethod
    def _from_oklab(cls, parent, oklab):
        """To Lab."""

        return oklab_to_oklch(oklab)

    @classmethod
    def _to_xyz(cls, parent, oklch):
        """To XYZ."""

        return Oklab._to_xyz(parent, cls._to_oklab(parent, oklch))

    @classmethod
    def _from_xyz(cls, parent, xyz):
        """From XYZ."""

        return cls._from_oklab(parent, Oklab._from_xyz(parent, xyz))



"""Pro Photo RGB color class."""
from ..spaces import RE_DEFAULT_MATCH
from .srgb import SRGB
from .xyz import XYZ
from .. import util
import re

ET = 1 / 512
ET2 = 16 / 512


def lin_prophoto_to_xyz(rgb):
    """
    Convert an array of linear-light prophoto-rgb values to CIE XYZ using  D50.D50.

    (so no chromatic adaptation needed afterwards)
    http://www.brucelindbloom.com/index.html?Eqn_RGB_XYZ_Matrix.html
    """

    m = [
        [7.9767494443060449e-01, 1.3519170147409817e-01, 3.1353354095297416e-02],
        [2.8804023786231026e-01, 7.1187409723579020e-01, 8.5664901899719714e-05],
        [0.0000000000000000e+00, 0.0000000000000000e+00, 8.2521000000000000e-01]
    ]

    return util.dot(m, rgb)


def xyz_to_lin_prophoto(xyz):
    """Convert XYZ to linear-light prophoto-rgb."""

    m = [
        [1.3459433009386652, -0.255607509316767, -0.051111765870885],
        [-0.544598869458717, 1.508167317720767, 0.0205351415866469],
        [0.0, 0.0, 1.2118127506937628]
    ]

    return util.dot(m, xyz)


def lin_prophoto(rgb):
    """
    Convert an array of prophoto-rgb values in the range 0.0 - 1.0 to linear light (un-corrected) form.

    Transfer curve is gamma 1.8 with a small linear portion.

    https://en.wikipedia.org/wiki/ProPhoto_RGB_color_space
    """

    result = []
    for i in rgb:
        # Mirror linear nature of algorithm on the negative axis
        if abs(i) < ET2:
            result.append(i / 16.0)
        else:
            result.append(util.npow(i, 1.8))
    return result


def gam_prophoto(rgb):
    """
    Convert an array of linear-light prophoto-rgb  in the range 0.0-1.0 to gamma corrected form.

    Transfer curve is gamma 1.8 with a small linear portion.

    https://en.wikipedia.org/wiki/ProPhoto_RGB_color_space
    """

    result = []
    for i in rgb:
        # Mirror linear nature of algorithm on the negative axis
        if abs(i) < ET:
            result.append(16.0 * i)
        else:
            result.append(util.npow(i, 1.0 / 1.8))
    return result


class ProPhotoRGB(SRGB):
    """Pro Photo RGB class."""

    SPACE = "prophoto-rgb"
    DEFAULT_MATCH = re.compile(RE_DEFAULT_MATCH.format(color_space=SPACE))
    WHITE = "D50"

    @classmethod
    def _to_xyz(cls, parent, rgb):
        """To XYZ."""

        return parent.chromatic_adaptation(cls.WHITE, XYZ.WHITE, lin_prophoto_to_xyz(lin_prophoto(rgb)))

    @classmethod
    def _from_xyz(cls, parent, xyz):
        """From XYZ."""

        return gam_prophoto(xyz_to_lin_prophoto(parent.chromatic_adaptation(XYZ.WHITE, cls.WHITE, xyz)))
